==> mono_repo-3.2.0.tar.gz: lib/src/commands/travis/travis_yaml.dart <==
// Copyright (c) 2020, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:collection';

import 'package:graphs/graphs.dart';
import 'package:path/path.dart' as p;

import '../../ci_shared.dart';
import '../../package_config.dart';
import '../../root_config.dart';
import '../../user_exception.dart';
import '../../yaml.dart';
import '../ci_script/generate.dart';

String generateTravisYml(
  RootConfig rootConfig,
  Map<String, String> commandsToKeys,
) {
  final orderedStages = _calculateOrderedStages(rootConfig);

  final jobs = rootConfig.expand((config) => config.jobs);

  var customTravis = '';
  if (rootConfig.monoConfig.travis.isNotEmpty) {
    customTravis = '\n# Custom configuration\n'
        '${toYaml(rootConfig.monoConfig.travis)}\n';
  }

  final branchConfig = rootConfig.monoConfig.travis.containsKey('branches')
      ? ''
      : '''
\n# Only building master means that we don't run two builds for each pull request.
${toYaml({
          'branches': {
            'only': ['master']
          }
        })}
''';

  int stageIndex(String value) => orderedStages.indexWhere((e) {
        if (e is String) {
          return e == value;
        }

        return (e as Map)['name'] == value;
      });

  final jobList = [
    ..._listJobs(jobs, commandsToKeys, rootConfig.monoConfig.mergeStages),
    if (rootConfig.monoConfig.selfValidateStage != null)
      _selfValidateTaskConfig(rootConfig.monoConfig.selfValidateStage),
  ]..sort((a, b) {
      var value = stageIndex(a['stage']).compareTo(stageIndex(b['stage']));

      for (var key in const ['env', 'script', 'dart', 'os']) {
        if (value != 0) {
          break;
        }
        final aVal = a[key];
        final bVal = b[key];
        if (aVal == null) {
          if (bVal == null) {
            value = 0;
          } else {
            // null (a) first
            value = -1;
          }
        } else {
          if (bVal == null) {
            // null (b) first
            value = 1;
          } else {
            value = aVal.compareTo(bVal);
          }
        }
      }
      return value;
    });

  return '''
$createdWith
${toYaml({'language': 'dart'})}
$customTravis
${toYaml({
    'jobs': {'include': jobList}
  })}

${toYaml({'stages': orderedStages})}
$branchConfig
${toYaml({
    'cache': {'directories': _cacheDirs(rootConfig)}
  })}
''';
}

Iterable<String> _cacheDirs(Iterable<PackageConfig> configs) {
  final items = SplayTreeSet<String>()..add('\$HOME/.pub-cache');

  for (var entry in configs) {
    for (var dir in entry.cacheDirectories) {
      items.add(p.posix.join(entry.relativePath, dir));
    }
  }

  return items;
}

/// Calculates the global stages ordering, and throws a [UserException] if it
/// detects any cycles.
List<Object> _calculateOrderedStages(RootConfig rootConfig) {
  // Convert the configs to a graph so we can run strongly connected components.
  final edges = <String, Set<String>>{};

  String previous;
  for (var stage in rootConfig.monoConfig.conditionalStages.keys) {
    edges.putIfAbsent(stage, () => <String>{});
    if (previous != null) {
      edges[previous].add(stage);
    }
    previous = stage;
  }

  final rootMentionedStages = <String>{
    ...rootConfig.monoConfig.conditionalStages.keys,
    ...rootConfig.monoConfig.mergeStages,
  };

  for (var config in rootConfig) {
    String previous;
    for (var stage in config.stageNames) {
      rootMentionedStages.remove(stage);
      edges.putIfAbsent(stage, () => <String>{});
      if (previous != null) {
        edges[previous].add(stage);
      }
      previous = stage;
    }
  }

  if (rootMentionedStages.isNotEmpty) {
    final items = rootMentionedStages.map((e) => '`$e`').join(', ');

    throw UserException(
      'Error parsing mono_repo.yaml',
      details: 'One or more stage was referenced in `mono_repo.yaml` that do '
          'not exist in any `mono_pkg.yaml` files: $items.',
    );
  }

  // Running strongly connected components lets us detect cycles (which aren't
  // allowed), and gives us the reverse order of what we ultimately want.
  final components = stronglyConnectedComponents(edges.keys, (n) => edges[n]);
  for (var component in components) {
    if (component.length > 1) {
      final items = component.map((e) => '`$e`').join(', ');
      throw UserException(
        'Not all packages agree on `stages` ordering, found '
        'a cycle between the following stages: $items.',
      );
    }
  }

  final orderedStages = components
      .map((c) {
        final stageName = c.first;

        final matchingStage =
            rootConfig.monoConfig.conditionalStages[stageName];

        return matchingStage?.toJson() ?? stageName;
      })
      .toList()
      .reversed
      .toList();

  if (rootConfig.monoConfig.selfValidateStage != null &&
      !orderedStages.contains(rootConfig.monoConfig.selfValidateStage)) {
    orderedStages.insert(0, rootConfig.monoConfig.selfValidateStage);
  }

  return orderedStages;
}

/// Lists all the jobs, setting their stage, environment, and script.
Iterable<Map<String, String>> _listJobs(
  Iterable<CIJob> jobs,
  Map<String, String> commandsToKeys,
  Set<String> mergeStages,
) sync* {
  final jobEntries = <CIJobEntry>[];

  for (var job in jobs) {
    final commands =
        job.tasks.map((task) => commandsToKeys[task.command]).toList();

    jobEntries.add(CIJobEntry(job, commands));
  }

  final groupedItems = groupCIJobEntries(jobEntries);

  for (var entry in groupedItems.entries) {
    final first = entry.value.first;
    if (mergeStages.contains(first.job.stageName)) {
      final packages = entry.value.map((t) => t.job.package).toList();
      yield first.jobYaml(packages);
    } else {
      yield* entry.value.map((jobEntry) => jobEntry.jobYaml());
    }
  }
}

extension on CIJobEntry {
  Map<String, String> jobYaml([List<String> packages]) {
    packages ??= [job.package];
    assert(packages.isNotEmpty);
    assert(packages.contains(job.package));

    return {
      'stage': job.stageName,
      'name': jobName(packages),
      'dart': job.sdk,
      'os': job.os,
      'env': 'PKGS="${packages.join(' ')}"',
      'script': '$ciScriptPath ${commands.join(' ')}',
    };
  }
}

Map<String, String> _selfValidateTaskConfig(String stageName) => {
      'stage': stageName,
      'name': selfValidateJobName,
      'os': 'linux',
      'script': selfValidateCommands.join(' && ')
    };

==> mono_repo-3.4.0.tar.gz: lib/src/commands/travis/travis_yaml.dart <==
// Copyright (c) 2020, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

import 'dart:collection';

import 'package:meta/meta.dart';
import 'package:path/path.dart' as p;

import '../../ci_shared.dart';
import '../../package_config.dart';
import '../../root_config.dart';
import '../../yaml.dart';
import '../ci_script/generate.dart';

String generateTravisYml(
  RootConfig rootConfig,
  Map<String, String> commandsToKeys,
) {
  final orderedStages = calculateOrderedStages(
      rootConfig, rootConfig.monoConfig.travisConditionalStages);

  final jobs = rootConfig.expand((config) => config.jobs);

  var customTravis = '';
  if (rootConfig.monoConfig.travis.isNotEmpty) {
    customTravis = '\n# Custom configuration\n'
        '${toYaml(rootConfig.monoConfig.travis)}\n';
  }

  final branchConfig = rootConfig.monoConfig.travis.containsKey('branches')
      ? ''
      : '''
\n# Only building master means that we don't run two builds for each pull request.
${toYaml({
          'branches': {
            'only': ['master']
          }
        })}
''';

  final jobList = [
    ..._listJobs(jobs, commandsToKeys, rootConfig.monoConfig.mergeStages),
    if (rootConfig.monoConfig.selfValidateStage != null)
      _selfValidateTaskConfig(rootConfig.monoConfig.selfValidateStage),
  ]..sort((a, b) {
      var value = orderedStages
          .indexOf(a['stage'])
          .compareTo(orderedStages.indexOf(b['stage']));

      for (var key in const ['env', 'script', 'dart', 'os']) {
        if (value != 0) {
          break;
        }
        final aVal = a[key];
        final bVal = b[key];
        if (aVal == null) {
          if (bVal == null) {
            value = 0;
          } else {
            // null (a) first
            value = -1;
          }
        } else {
          if (bVal == null) {
            // null (b) first
            value = 1;
          } else {
            value = aVal.compareTo(bVal);
          }
        }
      }
      return value;
    });

  final stageConfigs = orderedStages.map((stage) {
    final conditional = rootConfig.monoConfig.travisConditionalStages[stage];
    return conditional == null ? stage : conditional.toJson();
  }).toList();

  return '''
$createdWith
${toYaml({'language': 'dart'})}
$customTravis
${toYaml({
    'jobs': {'include': jobList}
  })}

${toYaml({'stages': stageConfigs})}
$branchConfig
${toYaml({
    'cache': {'directories': _cacheDirs(rootConfig)}
  })}
''';
}

Iterable<String> _cacheDirs(Iterable<PackageConfig> configs) {
  final items = SplayTreeSet<String>()..add('\$HOME/.pub-cache');

  for (var entry in configs) {
    for (var dir in entry.cacheDirectories) {
      items.add(p.posix.join(entry.relativePath, dir));
    }
  }

  return items;
}

/// Lists all the jobs, setting their stage, environment, and script.
Iterable<Map<String, String>> _listJobs(
  Iterable<CIJob> jobs,
  Map<String, String> commandsToKeys,
  Set<String> mergeStages,
) sync* {
  final jobEntries = <CIJobEntry>[];

  for (var job in jobs) {
    final commands =
        job.tasks.map((task) => commandsToKeys[task.command]).toList();

    jobEntries.add(CIJobEntry(job, commands));
  }

  final differentPackages = <String>{};
  final differentSdks = <String>{};

  for (var entry in jobEntries) {
    differentPackages.add(entry.job.package);
    differentSdks.add(entry.job.sdk);
  }

  final groupedItems = groupCIJobEntries(jobEntries);

  for (var entry in groupedItems.entries) {
    final first = entry.value.first;
    if (mergeStages.contains(first.job.stageName)) {
      final packages = entry.value.map((t) => t.job.package).toList();
      yield first.jobYaml(
        packages: packages,
        oneSdk: differentSdks.length == 1,
        onePackage: differentPackages.length == 1,
      );
    } else {
      yield* entry.value.map((jobEntry) => jobEntry.jobYaml(
            oneSdk: differentSdks.length == 1,
            onePackage: differentPackages.length == 1,
          ));
    }
  }
}

extension on CIJobEntry {
  Map<String, String> jobYaml({
    List<String> packages,
    @required bool oneSdk,
    @required bool onePackage,
  }) {
    packages ??= [job.package];
    assert(packages.isNotEmpty);
    assert(packages.contains(job.package));

    return {
      'stage': job.stageName,
      'name': jobName(
        packages,
        includeOs: true,
        includeSdk: oneSdk,
        includePackage: onePackage,
        includeStage: false,
      ),
      'dart': job.sdk,
      'os': job.os,
      'env': 'PKGS="${packages.join(' ')}"',
      'script': '$ciScriptPath ${commands.join(' ')}',
    };
  }
}

Map<String, String> _selfValidateTaskConfig(String stageName) => {
      'stage': stageName,
      'name': selfValidateJobName,
      'os': 'linux',
      'script': selfValidateCommands.join(' && ')
    };
